import madkit.kernel.*;
import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;

public class Simulation
{
    private Kernel theKernel;
    private TurtleEnvironment env;
    private EnvironnementPanel envPanel;
    private ArrayList preyList, predatorList;
    private Viewer viewer;
    private TurtleScheduler sched;
    private Counter counter;
    private PlotPanel pp;
    private int maxPredator, maxPrey;
    private boolean withImages;
    private int cellSize;

    public Simulation(int x, int y, int cellSize, int nbPrey,
		      int nbPredator, boolean withImages, Document worldDescriptionXml)
    {
	preyList     = new ArrayList();
	predatorList = new ArrayList();

	maxPredator = nbPredator + 100;
	maxPrey     = nbPrey + 100;

	System.out.println("[Simulation] cellSize " + cellSize);

	theKernel = new Kernel("simPredatorPrey");
	
	env = new TurtleEnvironment(x, y, "hunt");
	theKernel.launchAgent(env, "environnement", this, false);

	pp = new PlotPanel("Proies", x*cellSize+275, nbPrey+105);

	envPanel = new EnvironnementPanel(this, cellSize, withImages);
	this.withImages = withImages;
	this.cellSize   = cellSize;

	if(worldDescriptionXml != null)
	    {
		initXmlAnimals(worldDescriptionXml, x, y);
	    }
	else
	    {
		initRandomAnimals(nbPrey, nbPredator, x, y);
	    }

	viewer = new Viewer("hunt", envPanel, false);
	theKernel.launchAgent(viewer, "viewer", this, false);

	counter = new Counter("hunt", pp);
	theKernel.launchAgent(counter, "counter", this, false);

	sched = new TurtleScheduler("hunt");
	theKernel.launchAgent(sched, "scheduler", this, false);
    }

    private void initRandomAnimals(int nbPrey, int nbPredator, int x, int y)
    {
	for (int i = 0; i < nbPrey; i++)
	    {
		Prey prey = new Prey(env, 
				     (int) (Math.random()*(double)x), 
				     (int) (Math.random()*(double)y));
		preyList.add(prey);

		theKernel.launchAgent(prey, "bete", env, false);
	    }

	for (int j = 0; j < nbPredator; j++)
	    {
		Predator predator = new Predator(env, 
						 (int) (Math.random()*(double)x), 
						 (int) (Math.random()*(double)y));

		predatorList.add(predator);

		theKernel.launchAgent(predator, "bete", env, false);
	    }
    }

    private void initXmlAnimals(Document d, int sizex, int sizey)
    {	
	NodeList nl = d.getElementsByTagName("animal");

	for (int i = 0; i < nl.getLength() ; i++)
	    {
		NamedNodeMap nnm = nl.item(i).getAttributes();
		
		String type    = nnm.getNamedItem("type").getNodeValue();
		String xString = nnm.getNamedItem("x").getNodeValue();
		String yString = nnm.getNamedItem("y").getNodeValue();
		
		int x = (new Integer(xString)).intValue();
		int y = (new Integer(yString)).intValue();

		if(x > sizex || y > sizey)
		    {
			System.err.println("Erreur lors de l'ajout d'un animal de type " + type + " en (" + x + "," + y + ")");
			continue;
		    }

		if(type.compareTo("predator") == 0)
		    {
			Predator predator = new Predator(env, x, y);
			
			predatorList.add(predator);
			
			theKernel.launchAgent(predator, "bete", env, false);
		    }
		else if(type.compareTo("prey") == 0)
		    {
			Prey prey = new Prey(env, x, y);

			preyList.add(prey);
			
			theKernel.launchAgent(prey, "bete", env, false);			
		    }

		
	    }
    }

    public EnvironnementPanel getEnvironnementPanel()
    {
	return envPanel;
    }

    public PlotPanel getPlotPanel()
    {
	return pp;
    }

    public void destroy()
    {
	/** Remove the scheduler and the viewer */
	theKernel.killAgent(sched);
	theKernel.killAgent(viewer);
	theKernel.killAgent(counter);
	
	/** Remove all preys */
	ListIterator li = preyList.listIterator();

	while(li.hasNext())
	    {
		Prey p = (Prey) li.next();
		
		env.removeTurtle(p);
	    }

	/** Remove all predators */
	li = predatorList.listIterator();

	while(li.hasNext())
	    {
		Predator p = (Predator) li.next();
		
		env.removeTurtle(p);
	    }

	/** Finally, remove the environment */
	theKernel.killAgent(env);

	/** Remove all references */
	theKernel    = null;
	env          = null;
	preyList     = null;
	predatorList = null;
    }


    /**
     * Scheduler management functions
     */
    public void setSchedulerMode(int mode)
    {
	sched.setMode(mode);
    }

    public int getSchedulerMode()
    {
	return sched.getMode();
    }

    public void schedulerGiveOneMoreStep()
    {
	sched.giveOneMoreStep();
    }

    public void setSimulationSpeed(int speed)
    {
	sched.setSimulationSpeed(speed);
    }

    /**
     * Environnement management functions
     */
    public void setWrappingMode(boolean b)
    {
	env.setWrappingMode(b);
    }

    public int getWorldWidth()
    {
	return env.getWorldWidth();
    }

    public int getWorldHeight()
    {
	return env.getWorldHeight();
    }

    public boolean hasImages()
    {
	return withImages;
    }

    public int getCellSize()
    {
	return cellSize;
    }

    public List getPreyList()
    {
	return preyList;
    }
    
    public List getPredatorList()
    {
	return predatorList;
    }

    public boolean addTurtle(int x, int y, int type)
    {
	if(type == Turtle.PREY_TYPE)
	    {
		System.out.println(preyList.size() + " <=> " + maxPrey);

		if(preyList.size() == maxPrey)
		    {
			return false;
		    }

		Prey p = new Prey(env, x, y);
		env.addTurtle(p, x, y);
		preyList.add(p);
		theKernel.launchAgent(p, "bete", env, false);
	    }
	else if(type == Turtle.PREDATOR_TYPE)
	    {
		System.out.println(predatorList.size() + " <=> " + maxPredator);

		if(predatorList.size() == maxPredator)
		    {
			return false;
		    }

		Predator p = new Predator(env, x, y);
		env.addTurtle(p, x, y);
		predatorList.add(p);
		theKernel.launchAgent(p, "bete", env, false);
	    }

	return true;
    }

    public void delTurtle(int x, int y)
    {
	Turtle[] t = env.turtlesAt(x, y);
	
	if(t == null)
	    {
		return;
	    }

	if(t[0].getType() == Turtle.PREY_TYPE)
	    {
		preyList.remove(t[0]);
	    }
	else
	    {
		predatorList.remove(t[0]);
	    }

	env.removeTurtle(t[0]);
    }

    public boolean isTurtleAt(int x, int y)
    {
	Turtle t[] = env.turtlesAt(x, y);

	if(t.length == 0)
	    {
		return false;
	    }
	else 
	    {
		return true;
	    }
    }

    /**
     * Viewer management functions
     */
    public void setGraphicInfoMode(boolean b)
    {
	viewer.setGraphicInfoMode(b);
    }

    /**
     * Turtle management functions
     */
    public void setPredatorVisionRadius(int vr)
    {
	ListIterator li = predatorList.listIterator();

	while(li.hasNext())
	    {
		Predator p = (Predator) li.next();
		
		p.setVisionRadius(vr);
	    }	
    }

    public void setPreyVisionRadius(int vr)
    {
	ListIterator li = preyList.listIterator();

	while(li.hasNext())
	    {
		Prey p = (Prey) li.next();
		
		p.setVisionRadius(vr);
	    }	
    }

    public void setPreySpeed(int val)
    {
	ListIterator li = preyList.listIterator();

	while(li.hasNext())
	    {
		Prey p = (Prey) li.next();
		
		p.setSpeed(val);
	    }		
    }

    public void setPreyGroupingNb(int val)
    {
	ListIterator li = preyList.listIterator();

	while(li.hasNext())
	    {
		Prey p = (Prey) li.next();
		
		p.setNbParGroupe(val);
	    }		
    }

    public void setPredatorGroupingNb(int val)
    {
	ListIterator li = predatorList.listIterator();

	while(li.hasNext())
	    {
		Predator p = (Predator) li.next();
		
		p.setNbParGroupe(val);
	    }		
    }

    public void setPreyGrouping(boolean val)
    {
	ListIterator li = preyList.listIterator();

	while(li.hasNext())
	    {
		Prey p = (Prey) li.next();
		
		p.setRegroupement(val);
	    }		
    }
    
    public void setPredatorGrouping(boolean val)
    {
	ListIterator li = predatorList.listIterator();

	while(li.hasNext())
	    {
		Predator p = (Predator) li.next();
		
		p.setRegroupement(val);
	    }		
    }

    public void setPreyStrategy(int val)
    {
	ListIterator li = preyList.listIterator();

	while(li.hasNext())
	    {
		Prey p = (Prey) li.next();
		
		p.setStrategy(val);
	    }		
    }

    public void setPredatorStrategy(int val)
    {
	ListIterator li = predatorList.listIterator();

	while(li.hasNext())
	    {
		Predator p = (Predator) li.next();
		
		p.setStrategy(val);
	    }		
    }

}
